
#ifndef  DISABLE_ZLIB


/**
@brief   ZLIB圧縮用ライブラリ
@file    gzip_tool.c
@author  Fumi.Iseki (C)


@attention
このプログラムは zlib を使用しています．@n
*/


#include "gz_tool.h"



/*
int  gz_encode_file(const char* fmfn, const char* tofn)

	機能：ファイル fmfn を圧縮して gzip ファイル tofn へ保存する．

	戻り値：読み込んだファイルのサイズ．
*/
int  gz_encode_file(const char* fmfn, const char* tofn)
{
	int  cc, sz;
	unsigned char buf[RECVBUFSZ];
	gzFile gf;
	FILE*  fp;

	fp = fopen(fmfn, "rb");
	if (fp==NULL) return -1;

	gf = gzopen(tofn, "wb");
	if (gf==NULL) {
		fclose(fp);
		return -1;
	}

	memset(buf, 0, RECVBUFSZ);
	sz = cc = fread(buf, RECVBUFSZ, 1, fp);
	while(cc>0) {
		//gzread(gf, (voidp)buf, cc);
		gzwrite(gf, (voidp)buf, cc);
		memset(buf, 0, cc);
		cc = fread(buf, RECVBUFSZ, 1, fp);
		sz += cc;
	}

	gzclose(gf);	
	fclose(fp);

	return sz;
}




/**
int  gz_decode_file(const char* fmfn, const char* tofn)

	機能：gzip ファイル fmfn を解凍して ファイル tofn へ保存する．

	戻り値：書き込んだファイルのサイズ．
*/
int  gz_decode_file(const char* fmfn, const char* tofn)
{
	FILE* infp;
	FILE* otfp;

	infp = fopen(fmfn, "rb");
	if (infp==NULL) return -1;

	otfp = fopen(tofn, "wb");
	if (otfp==NULL) {
		fclose(infp);
		return -1;
	}

	int sz = gz_decode_fp(infp, otfp);

	fclose(infp);
	fclose(otfp);

	return sz;

/*
	int  cc, sz;
	unsigned char buf[RECVBUFSZ];
	gzFile gf;
	FILE*  fp;

	gf = gzopen(fmfn, "rb");
	if (gf==NULL) return -1;

	fp = fopen(tofn, "wb");
	if (fp==NULL) {
		gzclose(gf);
		return -1;
	}

	memset(buf, 0, RECVBUFSZ);
	sz = cc = gzread(gf, (voidp)buf, RECVBUFSZ);
	while(cc>0) {
		fwrite(buf, cc, 1, fp);
		memset(buf, 0, cc);
		cc = gzread(gf, (voidp)buf, RECVBUFSZ);
		sz += cc;
	}

	gzclose(gf);	
	fclose(fp);

	return sz;
*/
}




int  gz_decode_file_replace(const char* fn, const char* tempdir)
{
	int cc;
	char* tempfn;

	tempfn = temp_filename(tempdir, 16);
	cc = gz_decode_file(fn, tempfn);
	if (cc<=0) {
		unlink(tempfn);
		free(tempfn);
		return cc;
	}

	unlink(fn);
	rename(tempfn, fn);
	freeNull(tempfn);

	return cc;
} 





int  gz_decode_fp(FILE* infp, FILE* otfp)
{
	if (infp==NULL || otfp==NULL) return 0;
	
	int sz = 0;
	unsigned char inbuf[LBUF];
	unsigned char otbuf[BUFSZ];

	//
	z_stream zstrm;

	zstrm.zalloc 	= Z_NULL;
	zstrm.zfree  	= Z_NULL;
	zstrm.opaque 	= Z_NULL;
	zstrm.next_in  	= Z_NULL;
	zstrm.avail_in 	= 0;

	int ret = inflateInit2(&zstrm, 47);
	if (ret!=Z_OK) return -1;

	//
	do {
		zstrm.next_in  = inbuf;
		zstrm.avail_in = fread(inbuf, 1, LBUF, infp);

		do {
			zstrm.next_out  = otbuf;
			zstrm.avail_out = BUFSZ; 

			ret = inflate(&zstrm, Z_NO_FLUSH);
			if (ret!=Z_OK && ret!=Z_STREAM_END) {
				DEBUG_MODE print_message("GZ_DECODE_FP: ERROR: inflate error = %d\n", ret);
				inflateEnd(&zstrm);
				if (sz==0) sz = -1;
				return sz;
			}

			int otsz = BUFSZ - zstrm.avail_out; 
			fwrite(otbuf, otsz, 1, otfp);			
			sz += otsz;

      		if (ret==Z_STREAM_END && zstrm.avail_in!=0) {
        		ret = inflateReset(&zstrm);
			} 

		} while (ret!=Z_STREAM_END && zstrm.avail_in>0);
		//
	} while(!feof(infp));

	inflateEnd(&zstrm);

	return sz;
}




////////////////////////////////////////////////////////////////////////////////////////////////

/*
int  gz_encode_gzfp(FILE* fp, gzFile* gf)

	機能：ファイル fp を圧縮して gzip ファイル *gf へ保存する．

	戻り値：読み込んだファイルのサイズ．
*/
int  gz_encode_gzfp(FILE* fp, gzFile* gf)
{
	int  cc, sz;
	unsigned char buf[RECVBUFSZ];

	if (*gf==NULL) return -1;

	memset(buf, 0, RECVBUFSZ);
	sz = cc = fread(buf, RECVBUFSZ, 1, fp);
	while(cc>0) {
		//gzread(*gf, (voidp)buf, cc);
		gzwrite(*gf, (voidp)buf, cc);
		memset(buf, 0, cc);
		cc = fread(buf, RECVBUFSZ, 1, fp);
		sz += cc;
	}

	return sz;
}




/**
int  gz_decode_gzfp(gzFile* gf, FILE* fp)

	機能：gzip ファイル gf を解凍して ファイル fp へ保存する．

	戻り値：書き込んだファイルのサイズ．
*/
int  gz_decode_gzfp(gzFile* gf, FILE* fp)
{
	int  cc, sz;
	unsigned char buf[RECVBUFSZ];

	if (*gf==NULL) return -1;

	memset(buf, 0, RECVBUFSZ);
	sz = cc = gzread(*gf, (voidp)buf, RECVBUFSZ);
	while(cc>0) {
		fwrite(buf, cc, 1, fp);
		memset(buf, 0, cc);
		cc = gzread(*gf, (voidp)buf, RECVBUFSZ);
		sz += cc;
	}

	return sz;
}




#endif
